#ifndef CHROMEOS_CAMERA_LIB_MEDIAPIPE_SHIM_FRAMEWORK_CALCULATOR_FRAMEWORK_H_
#define CHROMEOS_CAMERA_LIB_MEDIAPIPE_SHIM_FRAMEWORK_CALCULATOR_FRAMEWORK_H_

#include <cstdint>
#include <functional>
#include <memory>
#include <string>
#include <type_traits>
#include <utility>

namespace mediapipe {
class CalculatorContract;
class CalculatorContext;
class Packet;
class PacketType;
}  // namespace mediapipe

namespace mediapipe_shim {

// Analogous to mediapipe::Packet, but unique.
class Packet {
 public:
  Packet();
  Packet(const Packet&) = delete;
  Packet& operator=(const Packet&) = delete;
  Packet(Packet&&);
  Packet& operator=(Packet&&);
  ~Packet();

  Packet At(int64_t timestamp_us) const&;
  Packet At(int64_t timestamp_us) &&;
  bool IsEmpty() const;
  template <typename T>
  const T& Get() const;
  // Returns nullptr if not the sole owner.
  template <typename T>
  std::unique_ptr<T> Consume();
  int64_t TimestampUs() const;
  std::string DebugString() const;

 private:
  friend class CalculatorContext;
  friend class CalculatorGraphImpl;
  template <typename T>
  friend Packet Adopt(T* ptr);

  explicit Packet(mediapipe::Packet packet);
  const mediapipe::Packet& internal() const;
  mediapipe::Packet into_internal() &&;

  std::unique_ptr<mediapipe::Packet> packet_;
};

// Analogous to mediapipe::Adopt.
template <typename T>
Packet Adopt(T* ptr);

// Analogous to mediapipe::AdoptAsUniquePtr.
template <typename T>
Packet AdoptAsUniquePtr(T* ptr) {
  return Adopt(new std::unique_ptr<T>(ptr));
}

// Analogous to mediapipe::MakePacket.
template <typename T, typename... Args>
Packet MakePacket(Args&&... args) {
  return Adopt(new T(std::forward<Args>(args)...));
}

// Analogous to mediapipe::GetFromUniquePtr.
template <typename T>
T* GetFromUniquePtr(const Packet& packet) {
  return packet.Get<std::unique_ptr<T>>().get();
}

// Analogous to mediapipe::PacketType.
class PacketType {
 public:
  template <typename T>
  PacketType& Set();
  PacketType& SetAny();
  PacketType& SetNone();
  PacketType& SetSameAs(const PacketType* type);
  PacketType& Optional();

 private:
  friend class CalculatorContract;

  explicit PacketType(mediapipe::PacketType* type);
  const mediapipe::PacketType* internal() const { return type_; }

  mediapipe::PacketType* type_ = nullptr;
};

enum class Port {
  kInput,
  kOutput,
  kSideInput,
  kSideOutput,
};

// Analogous to mediapipe::CalculatorContract.
class CalculatorContract {
 public:
  const std::string& GetNodeName() const;

  // Analogous to Inputs(), Outputs(), etc. provided functions.
  bool HasTag(Port port, const std::string& tag) const;
  int NumEntries(Port port, const std::string& tag) const;

  // Analogous to Inputs().Get(tag, index), Outputs().Get(tag, index), etc. The
  // returned object cannot outlive this object.
  PacketType Type(Port port, const std::string& tag, int index);

 private:
  friend class CalculatorBaseFactoryWrapper;
  friend class GlCalculatorHelper;

  explicit CalculatorContract(mediapipe::CalculatorContract* contract);
  mediapipe::CalculatorContract* internal() const { return contract_; }

  mediapipe::CalculatorContract* contract_ = nullptr;
};

// Analogous to mediapipe::CalculatorContext.
class CalculatorContext {
 public:
  const std::string& NodeName() const;
  int NodeId() const;
  const std::string& CalculatorType() const;
  int64_t InputTimestampUs() const;

  // Analogous to Inputs(), Outputs(), etc. provided functions.
  bool HasTag(Port port, const std::string& tag) const;
  int NumEntries(Port port, const std::string& tag) const;

  Packet GetInputSidePacket(const std::string& tag, int index) const;
  void SetOutputSidePacket(const std::string& tag, int index,
                           const Packet& packet);
  Packet GetInput(const std::string& tag, int index) const;
  Packet GetInput(const std::string& tag, int index);
  void AddOutput(const std::string& tag, int index, Packet packet);

 private:
  friend class CalculatorBaseFactoryWrapper;
  friend class CalculatorBaseWrapper;
  friend class GlCalculatorHelper;

  explicit CalculatorContext(mediapipe::CalculatorContext* context);
  mediapipe::CalculatorContext* internal() const { return context_; }

  mediapipe::CalculatorContext* context_ = nullptr;
};

// Analogous to mediapipe::CalculatorBase.
class CalculatorBase {
 public:
  virtual ~CalculatorBase() = default;
  virtual bool Open(CalculatorContext* context) = 0;
  virtual bool Process(CalculatorContext* context) = 0;
  virtual bool Close(CalculatorContext* context) = 0;
};

namespace registry_internal {

// Helper classes for calculator registration.
class CalculatorBaseFactory {
 public:
  virtual ~CalculatorBaseFactory() = default;
  virtual bool GetContract(CalculatorContract* contract) = 0;
  virtual std::unique_ptr<CalculatorBase> CreateCalculator(
      CalculatorContext* context) = 0;
};

template <typename T>
class CalculatorBaseFactoryFor : public CalculatorBaseFactory {
 public:
  static_assert(std::is_base_of<CalculatorBase, T>::value);

  bool GetContract(CalculatorContract* contract) override {
    return T::GetContract(contract);
  }

  std::unique_ptr<CalculatorBase> CreateCalculator(
      CalculatorContext* context) override {
    return std::make_unique<T>();
  }
};

}  // namespace registry_internal

// Analogous to mediapipe::CalculatorBaseRegistry.
class CalculatorBaseRegistry {
 public:
  // Register a calculator of type T to the MediaPipe framework. Expected to be
  // called only once for each calculator in the program.
  template <typename T>
  static void Register(const char* name) {
    RegisterInternal(
        name, std::make_unique<registry_internal::CalculatorBaseFactoryFor<T>>);
  }

  // Returns whether a calculator name is registered.
  static bool IsRegistered(const char* name);

 private:
  using Function = std::function<
      std::unique_ptr<registry_internal::CalculatorBaseFactory>()>;

  static void RegisterInternal(const char* name, Function func);
};

}  // namespace mediapipe_shim

#endif  // CHROMEOS_CAMERA_LIB_MEDIAPIPE_SHIM_FRAMEWORK_CALCULATOR_FRAMEWORK_H_
